Package: scoringRules
Type: Package
Title: Scoring Rules for Parametric and Simulated Distribution
        Forecasts
Version: 1.1.3
Date: 2024-09-12
Authors@R: c(person(given = c("Alexander", "I."), family = "Jordan", role = "aut",
                    comment = c(ORCID = "0000-0001-7423-1352")), 
             person(given = "Fabian", family = "Krueger", 
                    role = c("aut", "cre"), 
                    email = "Fabian.Krueger83@gmail.com",
                    comment = c(ORCID = "0000-0002-5112-9037")),
             person(given = "Sebastian", family = "Lerch", role = "aut",
                    comment = c(ORCID = "0000-0002-3467-4375")), 
             person(given = "Sam", family = "Allen", role = "aut",
                    comment = c(ORCID = "0000-0003-1971-8277")),
             person(given = "Maximiliane", family = "Graeter", role = "ctb"))
Description: Dictionary-like reference for computing scoring rules in a wide
    range of situations. Covers both parametric forecast distributions (such as
    mixtures of Gaussians) and distributions generated via simulation. Further 
    details can be found in the package vignettes <doi:10.18637/jss.v090.i12>, 
    <doi:10.18637/jss.v110.i08>.
URL: https://github.com/FK83/scoringRules
License: GPL (>= 2)
Imports: Rcpp (>= 0.12.0), methods, MASS, knitr
Depends: R (>= 3.00)
Suggests: gsl (>= 1.8-3), hypergeo(>= 1.0), rmarkdown, testthat, crch,
        ggplot2
LinkingTo: Rcpp, RcppArmadillo
RoxygenNote: 7.3.2
VignetteBuilder: knitr
Encoding: UTF-8
NeedsCompilation: yes
Packaged: 2024-09-12 12:03:27 UTC; fabian
Author: Alexander I. Jordan [aut] (<https://orcid.org/0000-0001-7423-1352>),
  Fabian Krueger [aut, cre] (<https://orcid.org/0000-0002-5112-9037>),
  Sebastian Lerch [aut] (<https://orcid.org/0000-0002-3467-4375>),
  Sam Allen [aut] (<https://orcid.org/0000-0003-1971-8277>),
  Maximiliane Graeter [ctb]
Maintainer: Fabian Krueger <Fabian.Krueger83@gmail.com>
Repository: CRAN
Date/Publication: 2024-09-18 14:20:02 UTC
Built: R 4.4.1; aarch64-apple-darwin22.6.0; 2024-09-20 17:47:51 UTC; unix
